#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-6-21 上午9:59
@File    : __init__.py.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
module_version = 'v3'


if __name__ == "__main__":
    pass
